//  ICE Revision: $Id$

#include "FieldValueExpressionDriver.H"
#include "FieldValuePluginFunction.H"

#include <Random.H>
#include <wallDist.H>
#include <nearWallDist.H>
#include <dimensionedVector.H>
#include "cellSet.H"
#include "faceSet.H"
#include "pointSet.H"

#include "addToRunTimeSelectionTable.H"

namespace Foam {

word FieldValueExpressionDriver::driverName_="internalField";

defineTypeNameAndDebug(FieldValueExpressionDriver, 0);

addNamedToRunTimeSelectionTable(CommonValueExpressionDriver, FieldValueExpressionDriver, dictionary, internalField);
addNamedToRunTimeSelectionTable(CommonValueExpressionDriver, FieldValueExpressionDriver, idName, internalField);

FieldValueExpressionDriver::FieldValueExpressionDriver (
    const string& time,
    const Time& runTime,
    const fvMesh &mesh,
    bool cacheReadFields,
    bool searchInMemory,
    bool searchOnDisc
)
    : CommonValueExpressionDriver(
        cacheReadFields,
        searchInMemory,
        searchOnDisc
    ),
      time_(time),
      mesh_(mesh),
      runTime_(runTime),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      isPointField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
}

FieldValueExpressionDriver::FieldValueExpressionDriver (
    const word &id,
    const fvMesh &mesh
)
    : CommonValueExpressionDriver(
        false,
        true,
        false
    ),
      time_(""),
      mesh_(mesh),
      runTime_(mesh.time()),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      isPointField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
}
FieldValueExpressionDriver::FieldValueExpressionDriver (
    const FieldValueExpressionDriver &orig
)
    : CommonValueExpressionDriver(orig),
      time_(orig.time_),
      mesh_(orig.mesh_),
      runTime_(orig.runTime_),
      typ_(orig.typ_),
      isLogical_(orig.isLogical_),
      isSurfaceField_(orig.isSurfaceField_),
      resultDimension_(orig.resultDimension_)
{
}

FieldValueExpressionDriver::FieldValueExpressionDriver (
    const fvMesh &mesh,
    bool cacheReadFields,
    bool searchInMemory,
    bool searchOnDisc
)
    : CommonValueExpressionDriver(
        cacheReadFields,
        searchInMemory,
        searchOnDisc
    ),
      time_(""),
      mesh_(mesh),
      runTime_(mesh.time()),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      isPointField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
}

FieldValueExpressionDriver::FieldValueExpressionDriver (
    const dictionary &dict,
    const fvMesh &mesh
)
    : CommonValueExpressionDriver(dict),
      time_(""),
      mesh_(mesh),
      runTime_(mesh.time()),
      typ_("nothing"),
      isLogical_(false),
      isSurfaceField_(false),
      isPointField_(false),
      resultDimension_(0,0,0,0,0,0,0)
{
    if(dict.found("dimensions")) {
        resultDimension_.reset(dimensionSet(dict.lookup("dimensions")));
    }
}

FieldValueExpressionDriver::~FieldValueExpressionDriver ()
{
}

void FieldValueExpressionDriver::readVariablesAndTables(const dictionary &dict)
{
    CommonValueExpressionDriver::readVariablesAndTables(dict);

    if(dict.found("dimensions")) {
        resultDimension_.reset(dimensionSet(dict.lookup("dimensions")));
    }
}

void FieldValueExpressionDriver::parseInternal(int startToken)
{
    parserField::FieldValueExpressionParser parser (scanner_,*this,startToken,0);
    parser.set_debug_level (traceParsing());
    parser.parse ();
}

bool FieldValueExpressionDriver::isCellSet(const string &name)
{
    if(getTypeOfSet(name)=="cellSet") {
        return true;
    } else {
        return false;
    }
}

bool FieldValueExpressionDriver::isCellZone(const string &name)
{
    if(mesh_.cellZones().findZoneID(name)>=0) {
        return true;
    } else {
        return false;
    }
}

bool FieldValueExpressionDriver::isFaceSet(const string &name)
{
    if(getTypeOfSet(name)=="faceSet") {
        return true;
    } else {
        return false;
    }
}

bool FieldValueExpressionDriver::isFaceZone(const string &name)
{
    if(mesh_.faceZones().findZoneID(name)>=0) {
        return true;
    } else {
        return false;
    }
}

bool FieldValueExpressionDriver::isPointSet(const string &name)
{
    if(getTypeOfSet(name)=="pointSet") {
        return true;
    } else {
        return false;
    }
}

bool FieldValueExpressionDriver::isPointZone(const string &name)
{
    if(mesh_.pointZones().findZoneID(name)>=0) {
        return true;
    } else {
        return false;
    }
}

volScalarField *FieldValueExpressionDriver::makeModuloField(
    const volScalarField &a,
    const volScalarField &b)
{
    volScalarField *result_=makeConstantField<volScalarField>(0.);

    forAll(*result_,cellI) {
        scalar val=fmod(a[cellI],b[cellI]);

        if(fabs(val)>(b[cellI]/2)) {
            if(val>0) {
                val-=b[cellI];
            } else {
                val+=b[cellI];
            }
        }

        (*result_)[cellI]=val;
    }

    result_->correctBoundaryConditions();

    return result_;
}

volScalarField *FieldValueExpressionDriver::makeRandomField(label seed)
{
    volScalarField *f=makeConstantField<volScalarField>(0.);

    autoPtr<scalarField> rField(CommonValueExpressionDriver::makeRandomField(seed));
    f->internalField()=rField();

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeCellIdField()
{
    volScalarField *f=makeConstantField<volScalarField>(0.);

    forAll(*f,cellI) {
        (*f)[cellI]=scalar(cellI);
    }

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeGaussRandomField(label seed)
{
    volScalarField *f=makeConstantField<volScalarField>(0.);

    autoPtr<scalarField> rField(CommonValueExpressionDriver::makeGaussRandomField(seed));
    f->internalField()=rField();

    f->correctBoundaryConditions();

    return f;
}

volVectorField *FieldValueExpressionDriver::makePositionField()
{
    volVectorField *f=new volVectorField(
        IOobject
        (
            "pos",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.C().dimensions());
    *f=mesh_.C();
    f->dimensions().reset(dimless);

    f->correctBoundaryConditions();

    return f;
}

pointVectorField *FieldValueExpressionDriver::makePointPositionField()
{
    pointVectorField *f=new pointVectorField(
        IOobject
        (
            "fpos",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        this->pMesh(),
        vector::zero,
        "zeroGradient"
    );
    f->internalField()=mesh_.points();

    return f;
}

surfaceVectorField *FieldValueExpressionDriver::makeFacePositionField()
{
    surfaceVectorField *f=new surfaceVectorField(
        IOobject
        (
            "fpos",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        vector::zero
    );
    f->dimensions().reset(mesh_.Cf().dimensions());
    *f=mesh_.Cf();
    f->dimensions().reset(dimless);

    return f;
}

surfaceVectorField *FieldValueExpressionDriver::makeFaceProjectionField()
{
    surfaceVectorField *f=new surfaceVectorField(
        IOobject
        (
            "fproj",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.Cf().dimensions());

    vector fmin(0,0,0);
    vector fmax(0,0,0);

    forAll(*f,faceI)
    {
        const face &fProp = mesh_.faces()[faceI];
        fmin = mesh_.points()[fProp[0]];
        fmax = fmin;
        forAll(fProp,pointI)
        {
            forAll(mesh_.points()[0],compI)
            {
                if(
                    mesh_.points()[fProp[pointI]].component(compI)
                    <
                    fmin.component(compI)
                ) {
                    fmin.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                }
                if(
                    mesh_.points()[fProp[pointI]].component(compI)
                    >
                    fmax.component(compI)
                ) {
                    fmax.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                }
            }
        }
        (*f)[faceI] = fmax - fmin;
    }
    forAll(mesh_.boundaryMesh(),patchI)
    {
        labelList cNumbers = mesh_.boundaryMesh()[patchI].faceCells();
        fvsPatchVectorField & fFace = f->boundaryField()[patchI];

        forAll(fFace,faceI)
        {
            const cell & cProp(mesh_.cells()[cNumbers[faceI]]);

            forAll(cProp,cJ)
            {
                const label patchID = mesh_.boundaryMesh().whichPatch(cProp[cJ]);

                if (patchID == patchI)
                {
                    const face & fProp = mesh_.faces()[cProp[cJ]];

                    fmin = mesh_.points()[fProp[0]];
                    fmax = fmin;

                    forAll(fProp,pointI)
                    {
                        forAll(mesh_.points()[0],compI)
                        {
                            if(
                                mesh_.points()[fProp[pointI]].component(compI)
                                <
                                fmin.component(compI)
                            ) {
                                fmin.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                            }
                            if(
                                mesh_.points()[fProp[pointI]].component(compI)
                                >
                                fmax.component(compI)
                            ) {
                                fmax.component(compI) = mesh_.points()[fProp[pointI]].component(compI);
                            }
                        }
                    }
                    fFace[faceI] = fmax - fmin;
                }
            }
        }
    }

    f->dimensions().reset(dimless);

    return f;
}

surfaceVectorField *FieldValueExpressionDriver::makeFaceField()
{
    surfaceVectorField *f=new surfaceVectorField(
        IOobject
        (
            "face",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        vector(0,0,0)
    );
    f->dimensions().reset(mesh_.Sf().dimensions());
    *f=mesh_.Sf();
    f->dimensions().reset(dimless);

    return f;
}

surfaceScalarField *FieldValueExpressionDriver::makeAreaField()
{
    surfaceScalarField *f=new surfaceScalarField(
        IOobject
        (
            "face",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        0.
    );
    f->dimensions().reset(mesh_.magSf().dimensions());
    *f=mesh_.magSf();
    f->dimensions().reset(dimless);

    return f;
}

volScalarField *FieldValueExpressionDriver::makeVolumeField()
{
    volScalarField *f=new volScalarField(
        IOobject
        (
            "vol",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        0.
    );
    const scalarField &V=mesh_.V();

    forAll(*f,cellI) {
        (*f)[cellI]=V[cellI];
    }

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeDistanceField()
{
    volScalarField *f=new volScalarField(
        IOobject
        (
            "dist",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        0.
    );
    f->dimensions().reset(mesh_.C().dimensions());
    wallDist dist(mesh_);
    *f=dist;
    f->dimensions().reset(dimless);

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeNearDistanceField()
{
    volScalarField *f=new volScalarField(
        IOobject
        (
            "dist",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        0.,
        "fixedValue"
    );
    f->dimensions().reset(mesh_.C().dimensions());
    nearWallDist dist(mesh_);
    f->boundaryField()==dist;
    f->dimensions().reset(dimless);

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeRDistanceField(const volVectorField& r)
{
    volScalarField *f=new volScalarField(
        IOobject
        (
            "rdist",
            time(),
            mesh_,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh_,
        0.
    );

    forAll(*f,cellI) {
        (*f)[cellI]=mag(mesh_.C()[cellI] - r[cellI]);
    }

    f->correctBoundaryConditions();

    return f;
}

volScalarField *FieldValueExpressionDriver::makeCellSetField(const string &name)
{
  volScalarField *f=makeConstantField<volScalarField>(0);

  IOobject head
      (
          name,
          time(),
          polyMesh::meshSubDir/"sets",
          mesh_,
          IOobject::MUST_READ,
          IOobject::NO_WRITE
      );

  if(!head.headerOk()) {;
      head=IOobject
          (
              name,
              "constant",
              polyMesh::meshSubDir/"sets",
              mesh_,
              IOobject::MUST_READ,
              IOobject::NO_WRITE
          );
      head.headerOk();
  }

  cellSet cs(head);
  labelList cells(cs.toc());

  forAll(cells,cellI) {
    (*f)[cells[cellI]]=1.;
  }

  f->correctBoundaryConditions();

  return f;
}

surfaceScalarField *FieldValueExpressionDriver::makeInternalFaceField()
{
    surfaceScalarField *f=makeConstantField<surfaceScalarField>(1,true);

    forAll(f->boundaryField(),patchI) {
        forAll((*f).boundaryField()[patchI],faceI) {
            f->boundaryField()[patchI][faceI]=0;
        }
    }

    return f;
}

surfaceScalarField *FieldValueExpressionDriver::makeOnPatchField(const string &name)
{
    surfaceScalarField *f=makeConstantField<surfaceScalarField>(0,true);

    label patchI=mesh().boundaryMesh().findPatchID(name);
    if(patchI<0) {
        FatalErrorIn("makeFaceSetField(const string &name)")
            << "Patch name " << name << " not in valid names"
                << mesh().boundaryMesh().names()
                << endl
                << exit(FatalError);
    } else {
        forAll((*f).boundaryField()[patchI],faceI) {
            (*f).boundaryField()[patchI][faceI]=1;
        }
    }

    return f;
}

surfaceScalarField *FieldValueExpressionDriver::makeFaceSetField(const string &name)
{
    surfaceScalarField *f=makeConstantField<surfaceScalarField>(0,true);

  IOobject head
      (
          name,
          time(),
          polyMesh::meshSubDir/"sets",
          mesh_,
          IOobject::MUST_READ,
          IOobject::NO_WRITE
      );

  if(!head.headerOk()) {;
      head=IOobject
          (
              name,
              "constant",
              polyMesh::meshSubDir/"sets",
              mesh_,
              IOobject::MUST_READ,
              IOobject::NO_WRITE
          );
      head.headerOk();
  }

  faceSet cs(head);
  labelList faces(cs.toc());

  forAll(faces,faceI) {
      if(faces[faceI] < mesh().nInternalFaces()) {
          (*f)[faces[faceI]]=1.;
      } else {
          label patchI=mesh().boundaryMesh().whichPatch(faces[faceI]);
          if(patchI<0) {
              FatalErrorIn("FieldValueExpressionDriver::makeFaceSetField(const string &name")
                  << "Face " << faces[faceI] << " of faceSet "
                      << name << " is not in the mesh"
                      << endl
                      << exit(FatalError);
          } else {
              (*f).boundaryField()[patchI][
                  faces[faceI]
                  -
                  mesh().boundaryMesh()[patchI].start()] = 1.;
          }
      }
  }

  return f;
}

pointScalarField *FieldValueExpressionDriver::makePointSetField(const string &name)
{
  pointScalarField *f=makePointConstantField<pointScalarField>(0);

  IOobject head
      (
          name,
          time(),
          polyMesh::meshSubDir/"sets",
          mesh_,
          IOobject::MUST_READ,
          IOobject::NO_WRITE
      );

  if(!head.headerOk()) {;
      head=IOobject
          (
              name,
              "constant",
              polyMesh::meshSubDir/"sets",
              mesh_,
              IOobject::MUST_READ,
              IOobject::NO_WRITE
          );
      head.headerOk();
  }

  pointSet cs(head);
  labelList points(cs.toc());

  forAll(points,pointI) {
    (*f)[points[pointI]]=1.;
  }

  return f;
}

volScalarField *FieldValueExpressionDriver::makeCellZoneField(const string &name)
{
  volScalarField *f=makeConstantField<volScalarField>(0);
  label zoneID=mesh_.cellZones().findZoneID(name);

  const cellZone &zone=mesh_.cellZones()[zoneID];

  forAll(zone,ind) {
      label cellI=zone[ind];
      (*f)[cellI]=1.;
  }

  f->correctBoundaryConditions();

  return f;
}

surfaceScalarField *FieldValueExpressionDriver::makeFaceZoneField(const string &name)
{
  surfaceScalarField *f=makeConstantField<surfaceScalarField>(0);
  label zoneID=mesh_.faceZones().findZoneID(name);

  const faceZone &zone=mesh_.faceZones()[zoneID];

  forAll(zone,ind) {
      label faceI=zone[ind];

      if(faceI < mesh().nInternalFaces()) {
          (*f)[faceI]=1.;
      } else {
          label patchI=mesh().boundaryMesh().whichPatch(faceI);
          if(patchI<0) {
              FatalErrorIn("FieldValueExpressionDriver::makeFaceZoneField(const string &name")
                  << "Face " << faceI << " of faceZone "
                      << name << " is not in the mesh"
                      << endl
                      << exit(FatalError);
          } else {
              (*f).boundaryField()[patchI][
                  faceI
                  -
                  mesh().boundaryMesh()[patchI].start()] = 1.;
          }
      }
  }

  return f;
}

pointScalarField *FieldValueExpressionDriver::makePointZoneField(const string &name)
{
  pointScalarField *f=makePointConstantField<pointScalarField>(0);
  label zoneID=mesh_.pointZones().findZoneID(name);

  const pointZone &zone=mesh_.pointZones()[zoneID];

  forAll(zone,ind) {
      label pointI=zone[ind];
      (*f)[pointI]=1.;
  }

  return f;
}

volVectorField *FieldValueExpressionDriver::makeVectorField
(
    volScalarField *x,
    volScalarField *y,
    volScalarField *z
) {
    volVectorField *f=makeConstantField<volVectorField>(vector(0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=vector((*x)[cellI],(*y)[cellI],(*z)[cellI]);
    }

    f->correctBoundaryConditions();

    return f;
}

volTensorField *FieldValueExpressionDriver::makeTensorField
(
    volScalarField *xx,volScalarField *xy,volScalarField *xz,
    volScalarField *yx,volScalarField *yy,volScalarField *yz,
    volScalarField *zx,volScalarField *zy,volScalarField *zz
) {
    volTensorField *f=makeConstantField<volTensorField>(tensor(0,0,0,0,0,0,0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=tensor(
            (*xx)[cellI],(*xy)[cellI],(*xz)[cellI],
            (*yx)[cellI],(*yy)[cellI],(*yz)[cellI],
            (*zx)[cellI],(*zy)[cellI],(*zz)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

volSymmTensorField *FieldValueExpressionDriver::makeSymmTensorField
(
    volScalarField *xx,volScalarField *xy,volScalarField *xz,
    volScalarField *yy,volScalarField *yz,
    volScalarField *zz
) {
    volSymmTensorField *f=makeConstantField<volSymmTensorField>(symmTensor(0,0,0,0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=symmTensor(
            (*xx)[cellI],(*xy)[cellI],(*xz)[cellI],
            (*yy)[cellI],(*yz)[cellI],
            (*zz)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

volSphericalTensorField *FieldValueExpressionDriver::makeSphericalTensorField
(
    volScalarField *xx
) {
    volSphericalTensorField *f=makeConstantField<volSphericalTensorField>(sphericalTensor(0));

    forAll(*f,cellI) {
        (*f)[cellI]=sphericalTensor(
            (*xx)[cellI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

surfaceVectorField *FieldValueExpressionDriver::makeSurfaceVectorField
(
    surfaceScalarField *x,
    surfaceScalarField *y,
    surfaceScalarField *z
)
{
    surfaceVectorField *f=makeConstantField<surfaceVectorField>(vector(0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=vector((*x)[faceI],(*y)[faceI],(*z)[faceI]);
    }

    return f;
}

surfaceTensorField *FieldValueExpressionDriver::makeSurfaceTensorField
(
    surfaceScalarField *xx,surfaceScalarField *xy,surfaceScalarField *xz,
    surfaceScalarField *yx,surfaceScalarField *yy,surfaceScalarField *yz,
    surfaceScalarField *zx,surfaceScalarField *zy,surfaceScalarField *zz
) {
    surfaceTensorField *f=makeConstantField<surfaceTensorField>(tensor(0,0,0,0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=tensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yx)[faceI],(*yy)[faceI],(*yz)[faceI],
            (*zx)[faceI],(*zy)[faceI],(*zz)[faceI]
        );
    }

    return f;
}

surfaceSymmTensorField *FieldValueExpressionDriver::makeSurfaceSymmTensorField
(
    surfaceScalarField *xx,surfaceScalarField *xy,surfaceScalarField *xz,
    surfaceScalarField *yy,surfaceScalarField *yz,
    surfaceScalarField *zz
) {
    surfaceSymmTensorField *f=makeConstantField<surfaceSymmTensorField>(symmTensor(0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=symmTensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yy)[faceI],(*yz)[faceI],
            (*zz)[faceI]
        );
    }

    return f;
}

surfaceSphericalTensorField *FieldValueExpressionDriver::makeSurfaceSphericalTensorField
(
    surfaceScalarField *xx
) {
    surfaceSphericalTensorField *f=makeConstantField<surfaceSphericalTensorField>(sphericalTensor(0));

    forAll(*f,faceI) {
        (*f)[faceI]=sphericalTensor(
            (*xx)[faceI]
        );
    }

    return f;
}

pointVectorField *FieldValueExpressionDriver::makePointVectorField
(
    pointScalarField *x,
    pointScalarField *y,
    pointScalarField *z
)
{
    pointVectorField *f=makePointConstantField<pointVectorField>(vector(0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=vector((*x)[faceI],(*y)[faceI],(*z)[faceI]);
    }

    f->correctBoundaryConditions();

    return f;
}

pointTensorField *FieldValueExpressionDriver::makePointTensorField
(
    pointScalarField *xx,pointScalarField *xy,pointScalarField *xz,
    pointScalarField *yx,pointScalarField *yy,pointScalarField *yz,
    pointScalarField *zx,pointScalarField *zy,pointScalarField *zz
) {
    pointTensorField *f=makePointConstantField<pointTensorField>(tensor(0,0,0,0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=tensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yx)[faceI],(*yy)[faceI],(*yz)[faceI],
            (*zx)[faceI],(*zy)[faceI],(*zz)[faceI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

pointSymmTensorField *FieldValueExpressionDriver::makePointSymmTensorField
(
    pointScalarField *xx,pointScalarField *xy,pointScalarField *xz,
    pointScalarField *yy,pointScalarField *yz,
    pointScalarField *zz
) {
    pointSymmTensorField *f=makePointConstantField<pointSymmTensorField>(symmTensor(0,0,0,0,0,0));

    forAll(*f,faceI) {
        (*f)[faceI]=symmTensor(
            (*xx)[faceI],(*xy)[faceI],(*xz)[faceI],
            (*yy)[faceI],(*yz)[faceI],
            (*zz)[faceI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

pointSphericalTensorField *FieldValueExpressionDriver::makePointSphericalTensorField
(
    pointScalarField *xx
) {
    pointSphericalTensorField *f=makePointConstantField<pointSphericalTensorField>(sphericalTensor(0));

    forAll(*f,faceI) {
        (*f)[faceI]=sphericalTensor(
            (*xx)[faceI]
        );
    }

    f->correctBoundaryConditions();

    return f;
}

const word FieldValueExpressionDriver::time() const
{
    if(time_!="") {
        return time_;
    } else {
        return CommonValueExpressionDriver::time();
    }
}

template<>
FieldValueExpressionDriver::SymbolTable<FieldValueExpressionDriver>::SymbolTable()
:
StartupSymbols()
{
    // default value
    insert("",parserField::FieldValueExpressionParser::token::START_DEFAULT);

    insert(
        "volScalarField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_SCALAR_COMMA
    );
    insert(
        "volScalarField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_SCALAR_CLOSE
    );
    insert(
        "volVectorField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_VECTOR_COMMA
    );
    insert(
        "volVectorField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_VECTOR_CLOSE
    );
    insert(
        "volTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_TENSOR_COMMA
    );
    insert(
        "volTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_TENSOR_CLOSE
    );
    insert(
        "volSymmTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_YTENSOR_COMMA
    );
    insert(
        "volSymmTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_YTENSOR_CLOSE
    );
    insert(
        "volSphericalTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_HTENSOR_COMMA
    );
    insert(
        "volSphericalTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_HTENSOR_CLOSE
    );
    insert(
        "volLogicalField_SC",
        parserField::FieldValueExpressionParser::token::START_VOL_LOGICAL_COMMA
    );
    insert(
        "volLogicalField_CL",
        parserField::FieldValueExpressionParser::token::START_VOL_LOGICAL_CLOSE
    );

    insert(
        "surfaceScalarField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_SCALAR_COMMA
    );
    insert(
        "surfaceScalarField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_SCALAR_CLOSE
    );
    insert(
        "surfaceVectorField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_VECTOR_COMMA
    );
    insert(
        "surfaceVectorField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_VECTOR_CLOSE
    );
    insert(
        "surfaceTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_TENSOR_COMMA
    );
    insert(
        "surfaceTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_TENSOR_CLOSE
    );
    insert(
        "surfaceSymmTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_YTENSOR_COMMA
    );
    insert(
        "surfaceSymmTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_YTENSOR_CLOSE
    );
    insert(
        "surfaceSphericalTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_HTENSOR_COMMA
    );
    insert(
        "surfaceSphericalTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_HTENSOR_CLOSE
    );
    insert(
        "surfaceLogicalField_SC",
        parserField::FieldValueExpressionParser::token::START_SURFACE_LOGICAL_COMMA
    );
    insert(
        "surfaceLogicalField_CL",
        parserField::FieldValueExpressionParser::token::START_SURFACE_LOGICAL_CLOSE
    );

    insert(
        "pointScalarField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_SCALAR_COMMA
    );
    insert(
        "pointScalarField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_SCALAR_CLOSE
    );
    insert(
        "pointVectorField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_VECTOR_COMMA
    );
    insert(
        "pointVectorField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_VECTOR_CLOSE
    );
    insert(
        "pointTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_TENSOR_COMMA
    );
    insert(
        "pointTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_TENSOR_CLOSE
    );
    insert(
        "pointSymmTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_YTENSOR_COMMA
    );
    insert(
        "pointSymmTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_YTENSOR_CLOSE
    );
    insert(
        "pointSphericalTensorField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_HTENSOR_COMMA
    );
    insert(
        "pointSphericalTensorField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_HTENSOR_CLOSE
    );
    insert(
        "pointLogicalField_SC",
        parserField::FieldValueExpressionParser::token::START_POINT_LOGICAL_COMMA
    );
    insert(
        "pointLogicalField_CL",
        parserField::FieldValueExpressionParser::token::START_POINT_LOGICAL_CLOSE
    );

    insert(
        "CL",
        parserField::FieldValueExpressionParser::token::START_CLOSE_ONLY
    );
    insert(
        "SC",
        parserField::FieldValueExpressionParser::token::START_COMMA_ONLY
    );
}

const FieldValueExpressionDriver::SymbolTable<FieldValueExpressionDriver> &FieldValueExpressionDriver::symbolTable()
{
    static SymbolTable<FieldValueExpressionDriver> actualTable;

    return actualTable;
}

int FieldValueExpressionDriver::startupSymbol(const word &name) {
    return symbolTable()[name];
}


autoPtr<CommonPluginFunction> FieldValueExpressionDriver::newPluginFunction(
    const word &name
) {
    return autoPtr<CommonPluginFunction>(
        FieldValuePluginFunction::New(
            *this,
            name
        ).ptr()
    );
}

bool FieldValueExpressionDriver::existsPluginFunction(
    const word &name
) {
    return FieldValuePluginFunction::exists(
        *this,
        name
    );
}

} // end namespace
